@import url("https://fonts.googleapis.com/css2?family=Cascadia+Mono:ital,wght@0,200..700;1,200..700&display=swap");

.sig + .sig {
  border-top-left-radius: 0;
  border-top-right-radius: 0;
  padding-top: 0.05rem;
}
.sig:has(+ .sig) {
  border-bottom-left-radius: 0;
  border-bottom-right-radius: 0;
  padding-bottom: 0.05rem;
}

body {
  --font-stack--monospace: "Cascadia Mono", "SFMono-Regular", Menlo, Consolas,
    Monaco, "Liberation Mono", "Lucida Console", monospace;
}

.sig:not(.sig-inline) {
  font-weight: 700;
}

code,
kbd,
pre,
samp {
  font-weight: 350;
}

pre.ascii-graphics {
  line-height: 1.3 !important;
}

.samp em::before {
  content: "{";
  font-style: normal;
  color: var(--color-foreground-muted);
}

.samp em::after {
  content: "}";
  font-style: normal;
  color: var(--color-foreground-muted);
}

.color-path-hl em::before {
  content: "{";
}

.color-path-hl em::after {
  content: "}";
}

.toc-tree a.reference code {
  white-space: nowrap;
}

.toc-scroll li {
  overflow: hidden;
  text-overflow: ellipsis;
  color: var(--color-toc-item-text);
}

.no-link-decorations a {
  color: var(--sd-color-card-text);
  text-decoration: none;
}

.api-ref-prefix {
  color: var(--color-foreground-muted);
}
